"""
Copyright (c) 2023, Florian GARDIN
All rights reserved.

This source code is licensed under the BSD-style license found in the
LICENSE file in the root directory of this source tree.
"""

import numpy as np
import mido
from fractions import Fraction as frac

import pandas as pd
from mido import MidiFile, MidiTrack, Message, MetaMessage
from ..constants import SILENCE, CONTINUATION, PITCH, TRACK, OFFSET, VELOCITY, DURATION, TEMPO, PEDAL
from .constants import OCTAVES
import os


def voice_to_channel(instrument_list, voice, instruments):
    """

    Parameters
    ----------
    instrument_list :
        
    voice :
        
    instruments :
        

    Returns
    -------

    """
    instrument_of_voice = instruments.get(voice, 0)
    return instrument_list.index(instrument_of_voice)



def matrix_to_events(matrix, 
                     output_file=None, 
                     ticks_per_beat=384,
                     tempo=120, 
                     instruments={}, 
                     time_signature=(4, 4),
                     instrument_names=None):
    matrix = np.asarray(matrix)
    events = [] # [(time, instrument, duration, message)]
    for el in matrix:
        pitch, offset, duration, velocity, track, silence, continuation, tempo, pedal = el
        # Add note on, note off for each track

    return events



def set_tracks(mid, nb_tracks, instruments, instrument_names, channels, tempo, time_signature, output_file=None):
    for i in range(nb_tracks + 1):
        track = MidiTrack()
        if (len(instrument_names) > i) and instrument_names[i].startswith('drum'):
            channels[i] = 9
            track.append(Message('program_change', program=0, time=0, channel=9))
        elif i in instruments.keys():
            track.append(Message('program_change', program=instruments.get(i, 0), time=0, channel=channels[i]))
        elif i != 9:
            track.append(Message('program_change', program=0, time=0, channel=channels[i]))
        mid.tracks.append(track)

    if isinstance(output_file, str):
        mid.tracks[0].append(MetaMessage("track_name", name=os.path.split(output_file)[1], time=int(0)))
    else:
        mid.tracks[0].append(MetaMessage("track_name", name='track', time=int(0)))
    mid.tracks[0].append(MetaMessage("set_tempo", tempo=mido.bpm2tempo(tempo), time=int(0)))
    mid.tracks[0].append(MetaMessage("time_signature", numerator=time_signature[0], denominator=time_signature[1], time=int(0)))

    return mid, channels


def setup_instruments(matrix, instrument_names, instruments):

    # Change instrument names == 'drums' to -1  in instrument names
    idxs_drums = [idx for idx, name in enumerate(instrument_names) if name.startswith('drums')]
    for idx in idxs_drums:
        instruments[idx] = -1


    instrument_names_new = []
    instrument_group = {}
    for track_nb, program in instruments.items():
        instrument_name = instrument_names[track_nb]
        if instrument_name.startswith('drums_'):
            instrument_group[-1] = instrument_group.get(-1, []) + [track_nb]
        else:
            instrument_group[program] = instrument_group.get(program, []) + [track_nb]

    # Change the tracks associated to each instrument
    instruments = {i: program for i, program in enumerate(instrument_group.keys())}
    invert_instruments = {program: i for i, program in instruments.items()}
    instruments = {key: val if val != -1 else 0 for key, val in instruments.items() }
    # Associate each track to the right new track
    for program, tracks in instrument_group.items():
        instrument_name = instrument_names[tracks[0]]
        if program == -1:
            instrument_name = 'drums_0'
        instrument_names_new.append(instrument_name)
        for i, track in enumerate(tracks):
            matrix[matrix[:, TRACK] == track, TRACK] = invert_instruments[program]

    instrument_names = instrument_names_new
    return instrument_names, matrix, instruments

def number_to_channel(n):
    """
    Given a number, return the corresponding channel (avoiding drum channel)
    Parameters
    ----------
    n :

    Returns
    -------

    """
    if n < 9:
        return n
    if n >= 9:
        return n + 1


def init_midi_file(matrix, instruments, ticks_per_beat=480, time_signature=(4, 4), anachrusis_time=0):
    mid = MidiFile()
    mid.ticks_per_beat = ticks_per_beat
    mid.type = 1

    bar_duration = time_signature[0] * 4 / time_signature[1]
    start_time = (-anachrusis_time) % bar_duration

    from fractions import Fraction as frac
    matrix[:, OFFSET] = matrix[:, OFFSET] + frac(start_time).limit_denominator(24)

    instrument_list = list(sorted(list(set([0] + list(instruments.values())))))
    nb_tracks = matrix[:, TRACK].max()
    channels = [number_to_channel(voice_to_channel(instrument_list, i, instruments)) for i in range(nb_tracks + 1)]

    return mid, nb_tracks, channels, matrix


def matrix_to_events(matrix):
    """
    Convert a matrix of notes, silences or continuations to a list of midi events (NOTE_ON, NOTE_OFF)
    It should :
    - Groupby each track
    - For each track sort by time
    - For each track, create a list of events (pitch, time, velocity, duration, track)
    - Use continuations to delay note_off event
    - Use silences to create note_off events
    - Use next note to create note_off events

    Parameters
    ----------
    matrix

    Returns
    -------
    events: np.ndarray
        List of events (pitch, time, velocity, duration, track)
    """
    pass


def merge_continuation_to_previous_note(df):
    """
    Modify the music dataframe by merging the duration of continuation notes into the previous note.
    Parameters
    ----------
    df : pandas.DataFrame
        DataFrame representing the music data with columns like PITCH, OFFSET, DURATION, etc.

    Returns
    -------
    pandas.DataFrame
        Updated DataFrame with merged durations for continuation notes.
    """
    from fractions import Fraction
    import pandas as pd

    # Initialize a dictionary to keep track of the last note index for each track
    last_note_index = {}

    for index, row in df.iterrows():
        track = row['TRACK']
        is_continuation = row['CONTINUATION']

        if is_continuation:
            # Check if there is a previous note in the same track to merge with
            if track in last_note_index:
                # Merge the duration with the previous note
                previous_note_index = last_note_index[track]
                df.at[previous_note_index, 'DURATION'] += row['DURATION']
            else:
                # If no previous note, this is an error case
                print(f"Error: Continuation note at index {index} has no preceding note in track {track}.")
        else:
            # Update the last note index for this track
            last_note_index[track] = index

    # Remove the continuation rows from the DataFrame
    df = df[(df['CONTINUATION'] == False) & (df['SILENCE'] == False)]

    return df


import numpy as np
import pandas as pd


def prepare_df_for_events(df):
    df = df.copy()
    df = df.sort_values(['TRACK', 'OFFSET'])

    def add_delta(df):
        df['DELTA'] = df['OFFSET'].diff().fillna(df['OFFSET'].iloc[0])
        return df

    df['EVENT_TYPE'] = 'NOTE_ON'
    df['INDEX'] = np.arange(len(df))
    df_copy = df.copy()
    df_copy['EVENT_TYPE'] = 'NOTE_OFF'
    df_copy['OFFSET'] = df_copy['OFFSET'] + df_copy['DURATION']

    df_events = pd.concat([df, df_copy], axis=0)
    df_events = df_events.sort_values(['TRACK', 'OFFSET', 'EVENT_TYPE'])
    df_events = df_events.groupby('TRACK', group_keys=False).apply(add_delta)
    df_events['PITCH'] = df_events['PITCH'] + 60

    return df_events[['EVENT_TYPE', 'OFFSET', 'PITCH', 'VELOCITY', 'DURATION', 'DELTA', 'TRACK', 'TEMPO', 'PEDAL']]


def apply_events(df_events, mid, channels, current_tempo):
    ticks_per_beat = mid.ticks_per_beat


    for idx, row in df_events.iterrows():
        vel = int(row['VELOCITY'])
        track_nb = int(row['TRACK'])
        event_type = row['EVENT_TYPE']
        delta = int(row['DELTA'] * ticks_per_beat)
        track = mid.tracks[track_nb]
        pitch = int(row['PITCH'])
        if event_type == 'NOTE_ON':
            track.append(Message('note_on', note=pitch, channel=channels[track_nb],
                                 velocity=vel, time=delta))
        elif event_type == 'NOTE_OFF':
            track.append(Message('note_off', note=pitch, channel=channels[track_nb],
                                 velocity=vel, time=delta))
        if row['TEMPO'] is not None and row['TEMPO'] == row['TEMPO'] and row['TEMPO'] != current_tempo:
            real_tempo = (480000 * 120) // int(row['TEMPO'])
            track.append(MetaMessage("set_tempo", tempo=real_tempo, time=int(0)))
            current_tempo = row['TEMPO']
        if row['PEDAL'] is not None and row['PEDAL']:
            track.append(Message('control_change', value=127, channel=channels[track_nb], control=4, time=int(0)))

    return mid

def matrix_to_pandas(matrix):
    pd.DataFrame(matrix)
    pass


def matrix_to_mid(matrix, output_file=None, ticks_per_beat=480, tempo=120, instruments={}, time_signature=(4, 4),
                  anachrusis_time=0,
                  instrument_names=None, one_track_per_instrument=True, **kwargs):
    """
    Convert a matrix of notes, silences or continuations to a midi file


    Parameters
    ----------
    anachrusis_time: float
        Time in seconds of the anachrusis
    instrument_names: list of str
        List of instrument names
    one_track_per_instrument: bool (default = True)
        If True, each instrument will be on a different track
    matrix :
        
    output_file :
         (Default value = None)
    ticks_per_beat :
         (Default value = 480)
    tempo :
         (Default value = 120)
    instruments :
         (Default value = {})
    **kwargs :

    Returns
    -------

    """
    if instrument_names is None:
        instrument_names = []



    df = pd.DataFrame(matrix, columns=['PITCH', 'OFFSET', 'DURATION',
                                       'VELOCITY', 'TRACK', 'SILENCE', 'CONTINUATION',
                                       'TEMPO', 'PEDAL'
                                       ])
    df_old = df.copy()

    df = merge_continuation_to_previous_note(df.copy())
    matrix = df.values


    #matrix = np.asarray(matrix)

    if one_track_per_instrument:
        instrument_names, matrix, instruments = setup_instruments(matrix, instrument_names, instruments)

    df = pd.DataFrame(matrix, columns=['PITCH', 'OFFSET', 'DURATION', 'VELOCITY', 'TRACK', 'SILENCE', 'CONTINUATION',
                                       'TEMPO', 'PEDAL'
                                       ])
    df_events = prepare_df_for_events(df)



    mid, nb_tracks, channels, matrix = init_midi_file(matrix, instruments, ticks_per_beat=480, time_signature=(4, 4), anachrusis_time=0)
    mid, channels = set_tracks(mid, nb_tracks, instruments, instrument_names, channels, tempo, time_signature, output_file=None)

    mid = apply_events(df_events, mid, channels, tempo)
    # sort_events = []
    # last_silence = True
    #
    # for row in matrix:
    #     pitch = row[PITCH] + 60 + 12 * OCTAVES.get(instruments.get(row[TRACK], 0), 0)
    #     row = row.tolist()
    #     if row[SILENCE]:
    #         last_silence = True
    #         sort_events.append([0, 2, row[OFFSET], row[TEMPO], row[PEDAL], row[TRACK]])
    #         continue
    #
    #     if row[CONTINUATION]:
    #         sort_events.append([0, 2, row[OFFSET], row[TEMPO], row[PEDAL], row[TRACK]])
    #
    #     if row[CONTINUATION] == 0:
    #         sort_events.append([pitch,
    #                             1, row[OFFSET], row[VELOCITY], row[TRACK], row[TEMPO], row[PEDAL]])
    #
    #     if not (row[CONTINUATION] and last_silence):
    #         sort_events.append([pitch, 0, (row[OFFSET] + row[DURATION]), row[TRACK], row[CONTINUATION],
    #                             row[TEMPO], row[PEDAL]
    #                             ])
    #         last_silence = False
    #
    # sort_events.sort(key=lambda tup: tup[2])
    # current_tempo = tempo
    # lapso = {idx: 0 for idx in range(nb_tracks + 1)}
    # prev_pitch = {idx: 0 for idx in range(nb_tracks + 1)}
    # for evt in sort_events:
    #     if evt[1] == 1:
    #         vel = int(evt[3])
    #         track_nb = int(evt[4])
    #         track = mid.tracks[track_nb]
    #         if evt[0] > 100 or evt[0] < 10:
    #             evt[0] = 0
    #             vel = 0
    #
    #         real_time = int((evt[2] - lapso[track_nb]) * ticks_per_beat)
    #
    #         track.append(Message('note_on', note=int(evt[0]), channel=channels[track_nb],
    #                              velocity=vel, time=real_time))
    #
    #         if evt[5] is not None and evt[5] != current_tempo:
    #             real_tempo =(480000 * 120) // evt[5]
    #             track.append(MetaMessage("set_tempo", tempo=real_tempo, time=int(0)))
    #             current_tempo = evt[5]
    #
    #         if evt[6] is not None and evt[6]:
    #             track.append(Message('control_change', value=127, channel=channels[track_nb], control=4, time=int(0)))
    #
    #         if evt[6] is not None and not evt[6]:
    #             track.append(Message('control_change', value=0, channel=channels[track_nb], control=4, time=int(0)))
    #
    #         lapso[track_nb] = evt[2]
    #         prev_pitch[track_nb] = int(evt[0])
    #     elif evt[1] == 0:
    #         track_nb = int(evt[3])
    #         track = mid.tracks[track_nb]
    #         if evt[0] > 100 or evt[0] < 10:
    #             evt[0] = 0
    #         if evt[4] == 1 and prev_pitch[track_nb] > 0:
    #             track[-1] = Message('note_off', note=prev_pitch[track_nb], channel=channels[track_nb],
    #                                 velocity=0, time=int((evt[2] - lapso[track_nb]) * ticks_per_beat + track[-1].time))
    #         else:
    #             track.append(Message('note_off', note=int(evt[0]), channel=channels[track_nb],
    #                                  velocity=0, time=int((evt[2] - lapso[track_nb]) * ticks_per_beat)))
    #         lapso[track_nb] = evt[2]
    #
    #     elif evt[1] == 2:
    #         track_nb = int(evt[5])
    #         track = mid.tracks[track_nb]
    #         tempo_change = evt[3]
    #         pedal_change = evt[4]
    #
    #         if tempo_change is not None and tempo_change != current_tempo:
    #             real_tempo = (480000 * 120) // tempo_change
    #             track.append(MetaMessage("set_tempo", tempo=real_tempo, time=int(0)))
    #             current_tempo = tempo_change
    #
    #         if pedal_change is not None and pedal_change:
    #             track.append(Message('control_change', value=127, channel=channels[track_nb], control=4, time=int(0)))
    #
    #         if pedal_change is not None and not pedal_change:
    #             track.append(Message('control_change', value=0, channel=channels[track_nb], control=4, time=int(0)))


    if output_file is not None:
        for track in mid.tracks:
            track.append(MetaMessage('end_of_track', time=(int(0))))
        if isinstance(output_file, str):
            mid.save(output_file)
        else:
            mid.save(file=output_file)

    return df_events





from scipy.optimize import linear_sum_assignment
import numpy as np


def modulo_distance(a, b, modulo=12):
    """Calculate the distance in modulo space."""
    return min(abs(a - b), modulo - abs(a - b))


def optimal_assignment(A, B):
    """Find the optimal assignment between elements of A and B to minimize the sum of distances."""
    # Create a cost matrix with dimensions len(A) x len(B)
    cost_matrix = np.zeros((len(A), len(B)))

    # Fill the cost matrix with distances
    for i, a_elem in enumerate(A):
        for j, b_elem in enumerate(B):
            cost_matrix[i, j] = modulo_distance(a_elem, b_elem)

    # Use linear_sum_assignment to find the optimal assignment
    row_ind, col_ind = linear_sum_assignment(cost_matrix)

    # Calculate the total cost of the optimal assignment
    total_cost = cost_matrix[row_ind, col_ind].sum()

    return (row_ind, col_ind), total_cost


def optimal_assignment_extended(A, B):
    """Find the optimal assignment between elements of A and every element of B."""
    # Number of times we need to duplicate the columns of the cost matrix
    duplication_factor = len(A) // len(B) + (1 if len(A) % len(B) > 0 else 0)

    # Create an expanded cost matrix
    expanded_cost_matrix = np.zeros((len(A), len(B) * duplication_factor))

    # Fill the expanded cost matrix with distances
    for i, a_elem in enumerate(A):
        for j, b_elem in enumerate(B * duplication_factor):
            expanded_cost_matrix[i, j] = modulo_distance(a_elem, B[j % len(B)])

    # Use linear_sum_assignment to find the optimal assignment
    row_ind, col_ind = linear_sum_assignment(expanded_cost_matrix)

    # Map the results back to the original elements of B
    original_col_ind = [index % len(B) for index in col_ind]

    # Calculate the total cost of the optimal assignment
    total_cost = expanded_cost_matrix[row_ind, col_ind].sum()

    return list(zip(row_ind, original_col_ind)), total_cost


def project_to_modulo_space(values, modulo=12):
    """Project the given values to their classes in Z/moduloZ."""
    return [value % modulo for value in values]


def find_nearest_value(original_value, assigned_class, modulo=12):
    """Find the nearest value in Z that is in the assigned class of Z/moduloZ."""
    base_value = original_value - original_value % modulo + assigned_class
    if original_value % modulo > modulo / 2:
        base_value += modulo
    return base_value


def optimal_assignment_with_projection(A, B):
    # Project A to Z/12Z
    projected_A = project_to_modulo_space(A)

    # Apply the extended assignment algorithm
    assignment, _ = optimal_assignment_extended(projected_A, B)
    # Map back to the nearest values in Z
    final_assignment = [(A[i], find_nearest_value(A[i], B[j])) for i, j in assignment]

    return final_assignment


def plot_chroma(chroma, t_start=None, t_end=None):
    import matplotlib.pyplot as plt
    import matplotlib.patches as patches
    import numpy as np

    Cs = chroma.to_numpy()
    t = chroma.columns.to_numpy().astype(float)

    fig, ax = plt.subplots()

    # Calculate the differences in the time vector for widths
    t_diff = np.diff(t, append=t[-1] + 1)

    # Plot each rectangle
    for i in range(Cs.shape[0]):
        for j in range(Cs.shape[1]):
            if Cs[i, j] == 1:
                rect = patches.Rectangle((t[j], i), t_diff[j], 1, linewidth=0, edgecolor='none', facecolor='black')
                ax.add_patch(rect)

    # Set the limits and aspect ratio
    ax.set_xlim([t[0], t[-1] + 1])
    ax.set_ylim([0, Cs.shape[0]])
    ax.set_aspect('auto')
    plt.grid()
    plt.show()


"""
Transform functions from musiclang to chromagrams and inverse
"""

import pandas as pd
def musiclang_to_chromagram(score, quantization=None):
    """
    Transform a musiclang score to a chromagram

    Parameters
    ----------
    score: musiclang.Score
    quantization: frac or None : Fraction of a beat to quantize to

    Returns
    -------
    Cs : chroma matrix
    t : time vector
    time_signature : time signature of the score

    """
    def chroma_matrix(A):
        A['pitch_class'] = pd.Categorical(A['pitch_class'], categories=list(range(12)))
        B = pd.crosstab(A['pitch_class'], A['time'], dropna=False)

        # Replace NaNs with 0s and convert to integers
        B = (B.fillna(0).astype(int) > 0).astype(int)

        # Sort the columns if needed
        B = B.sort_index(axis=1)
        return B

    time_signature = score.config['time_signature']
    seq = score.to_sequence()

    seq = seq[~seq['instrument'].str.startswith('drums')]
    seq = seq[seq['pitch'].notnull()]
    seq['pitch_class'] = seq['pitch'] % 12
    # Step 1: Create a DataFrame of time ranges
    if quantization is None:
        start_times = seq['start'].unique()
    else:
        start_times = np.arange(0, seq['end'].max(), quantization)

    times_df = pd.DataFrame(start_times, columns=['time']).merge(seq, how='cross')
    times_df = times_df[(times_df['start'] <= times_df['time']) &
                        (times_df['end'] > times_df['time'])
                        ]
    times_df = times_df[['time', 'pitch_class']]
    # Create a chromagram
    chroma = chroma_matrix(times_df)

    Cs = chroma.to_numpy().T
    t = chroma.columns.to_numpy().astype(float)

    return Cs, t, time_signature


def chroma_to_musiclang(Cs, t):
    """
    Transform a chromagram to a musiclang score

    Parameters
    ----------
    chroma

    Returns
    -------

    """
    from musiclang import Score
    from musiclang.library import I


    # Calculate the differences in the time vector for widths
    t_diff = np.diff(t, append=t[-1] + 1)

    chord = (I % I.M)

    score = []
    for idx, duration in enumerate(t_diff):
        chromas = np.argwhere(Cs[idx])[:, 0]
        chord_pitches = [chord.parse(chroma) for chroma in chromas]
        score.append(chord(*chord_pitches).set_duration(duration))

    return Score(score)

def chromagram_to_musiclang(Cs, t, base_voicing, time_signature, repeat_notes=False, reparse=True):
    """
    Transform a chromagram to a musiclang score using the base_voicing as a starting point

    Parameters
    ----------
    chroma: (Cs, t) tuple : Cs is the chroma matrix, t is the time vector
    base_voicing : list of pitches to use as a starting point

    Returns
    -------
    score : musiclang.Score
    """
    from musiclang.library import I
    from musiclang import Score
    from musiclang.transform.melody import ContinuationWhenSameNote

    t_diff = np.diff(t, append=t[-1] + 1)

    chord = (I % I.M)

    base_octaves = [b // 12 for b in base_voicing]
    score = []
    for idx, duration in enumerate(t_diff):
        chromas = np.argwhere(Cs[idx])[:, 0].tolist()
        assignment = optimal_assignment_with_projection(base_voicing, list(chromas))

        # Change the octave to limit to +1, -1 relative to base

        result = [l[1] for l in assignment]

        result_octaves = [l // 12 for l in result]
        for idx, (base_octave, result_octave) in enumerate(zip(base_octaves, result_octaves)):
            if base_octave - result_octave > 1:
                result[idx] += 12
            elif result_octave - base_octave > 1:
                result[idx] -= 12

        base_voicing = result

        chord_pitches = [chord.parse(chroma) for chroma in result]
        score.append(chord(*chord_pitches).set_duration(duration))


    score = Score(score, time_signature=time_signature)

    if not repeat_notes:
        score = ContinuationWhenSameNote()(score)

    if reparse:
        score = score.reparse()


    return score






from musiclang.library import *
from musiclang import Score


def test_to_pattern():
    # 1. Create the base score (See previous example)
    flute_melody = (s0 + s2 + s4 + s3 + s4 + s3 + s2 + s4).e  # We will need it later

    chord = (I % I.M)(
        harpsichord__0=(s0 + s2 + s4 + s2).o(-1),
        flute__0=flute_melody,
    )
    pattern = chord.to_pattern()

    assert pattern[0]['pattern'] == (x0 + bu1 + bu1 + bd1)
    assert pattern[0]['instrument'] == 'harpsichord'
    assert pattern[0]['note'] == 'lowest_note'


from miditok import REMI
from miditok.classes import TokenizerConfig
from miditoolkit import MidiFile




def test_exotic_time_signature():
    config = TokenizerConfig(
        beat_res={(0, 8): 16, (8, 16): 16},
        use_tempos=True,
        use_time_signatures=True,
        time_signature_range={8: [3, 12, 6, 9, 10, 11],
                              4: [5, 6, 7, 3, 2, 1, 4],
                              16: [3, 6, 7, 9, 12, 14, 15, 17],
                              2: [1, 2, 3, 4],
                              1: [1, 2, 3, 4]
                              },
        one_token_stream_for_programs=True,
        use_programs=True)

    tokenizer = REMI(
        tokenizer_config=config,
    )
    tokenizer.one_token_stream = True
    tokens = ['Bar_None', 'TimeSig_9/8', 'Position_0', 'Program_24', 'Pitch_52', 'Velocity_127',
              'Duration_0.8.16', 'Program_24', 'Pitch_64', 'Velocity_127', 'Duration_0.8.16',
              'Position_8', 'Program_24', 'Pitch_50', 'Velocity_127', 'Duration_0.8.16',
              'Program_24', 'Pitch_62', 'Velocity_127', 'Duration_0.8.16', 'Position_16',
              'Program_24', 'Pitch_49', 'Velocity_127', 'Duration_0.8.16', 'Program_24',
              'Pitch_61', 'Velocity_127', 'Duration_0.8.16', 'Position_24', 'Program_24',
              'Pitch_50', 'Velocity_127', 'Duration_0.8.16', 'Program_24', 'Pitch_62',
              'Velocity_127', 'Duration_0.8.16', 'Position_32', 'Program_24',
              'Pitch_40', 'Velocity_127', 'Duration_1.0.16', 'Program_24', 'Pitch_52',
              'Velocity_127', 'Duration_1.0.16', 'Program_24', 'Pitch_64', 'Velocity_127', 'Duration_1.0.16',

              'Bar_None', 'TimeSig_9/8', 'Position_0', 'Program_24', 'Pitch_52', 'Velocity_127',
              'Duration_0.8.16', 'Program_24', 'Pitch_64', 'Velocity_127', 'Duration_0.8.16',
              'Position_8', 'Program_24', 'Pitch_50', 'Velocity_127', 'Duration_0.8.16',
              'Program_24', 'Pitch_62', 'Velocity_127', 'Duration_0.8.16', 'Position_16',
              'Program_24', 'Pitch_49', 'Velocity_127', 'Duration_0.8.16', 'Program_24',
              'Pitch_61', 'Velocity_127', 'Duration_0.8.16', 'Position_24', 'Program_24',
              'Pitch_50', 'Velocity_127', 'Duration_0.8.16', 'Program_24', 'Pitch_62',
              'Velocity_127', 'Duration_0.8.16', 'Position_32', 'Program_24',
              'Pitch_40', 'Velocity_127', 'Duration_1.0.16', 'Program_24', 'Pitch_52',
              'Velocity_127', 'Duration_1.0.16', 'Program_24', 'Pitch_64', 'Velocity_127', 'Duration_1.0.16',

              ]

    idxs = [tokenizer[t] for t in tokens]
    midi = tokenizer.tokens_to_midi(idxs)
    import tempfile

    from fractions import Fraction
    # Save midi to a tempfile
    with tempfile.NamedTemporaryFile(suffix='.mid') as f:
        output_path = f.name
        midi.dump(output_path)
        # Reload with musiclang
        from musiclang import Score
        score = Score.from_midi(output_path, tokenize_before=True)
        assert [c.duration for c in score] == [Fraction(9, 2),
                                               Fraction(9, 2),
                                               ]



def test_time_signature_change():
    """
    Test musiclang parser capability to parse a midi file with various time signature changes

    Returns
    -------

    """
    config = TokenizerConfig(
        beat_res={(0, 8): 16, (8, 16): 16},
        use_tempos=True,
        use_time_signatures=True,
        time_signature_range={8: [3, 12, 6, 9, 10, 11],
                              4: [5, 6, 7, 3, 2, 1, 4],
                              16: [3, 6, 7, 9, 12, 14, 15, 17],
                              2: [1, 2, 3, 4],
                              1: [1, 2, 3, 4]
                              },
        one_token_stream_for_programs=True,
        use_programs=True)

    tokenizer = REMI(
        tokenizer_config=config,
    )
    tokenizer.one_token_stream = True
    tokens = ['Bar_None', 'TimeSig_4/4', 'Position_0', 'Program_24', 'Pitch_52', 'Velocity_127',
              'Duration_0.8.16', 'Program_24', 'Pitch_64', 'Velocity_127', 'Duration_0.8.16',
              'Position_8', 'Program_24', 'Pitch_50', 'Velocity_127', 'Duration_0.8.16',
              'Program_24', 'Pitch_62', 'Velocity_127', 'Duration_0.8.16', 'Position_16',
              'Program_24', 'Pitch_49', 'Velocity_127', 'Duration_0.8.16', 'Program_24',
              'Pitch_61', 'Velocity_127', 'Duration_0.8.16', 'Position_24', 'Program_24',
              'Pitch_50', 'Velocity_127', 'Duration_0.8.16', 'Program_24', 'Pitch_62',
              'Velocity_127', 'Duration_0.8.16', 'Position_32', 'Program_24',
              'Pitch_40', 'Velocity_127', 'Duration_1.0.16', 'Program_24', 'Pitch_52',
              'Velocity_127', 'Duration_1.0.16', 'Program_24', 'Pitch_64', 'Velocity_127', 'Duration_1.0.16',

              'Bar_None', 'TimeSig_6/8', 'Position_0', 'Program_24', 'Pitch_52', 'Velocity_127',
              'Duration_0.8.16', 'Program_24', 'Pitch_64', 'Velocity_127', 'Duration_0.8.16',
              'Position_8', 'Program_24', 'Pitch_50', 'Velocity_127', 'Duration_0.8.16',
              'Program_24', 'Pitch_62', 'Velocity_127', 'Duration_0.8.16', 'Position_16',
              'Program_24', 'Pitch_49', 'Velocity_127', 'Duration_0.8.16', 'Program_24',
              'Pitch_61', 'Velocity_127', 'Duration_0.8.16', 'Position_24', 'Program_24',
              'Pitch_50', 'Velocity_127', 'Duration_0.8.16', 'Program_24', 'Pitch_62',
              'Velocity_127', 'Duration_0.8.16', 'Position_32', 'Program_24',
              'Pitch_40', 'Velocity_127', 'Duration_1.0.16', 'Program_24', 'Pitch_52',
              'Velocity_127', 'Duration_1.0.16', 'Program_24', 'Pitch_64', 'Velocity_127', 'Duration_1.0.16',

              'Bar_None', 'TimeSig_4/4', 'Position_0', 'Program_24', 'Pitch_52', 'Velocity_127',
              'Duration_0.8.16', 'Program_24', 'Pitch_64', 'Velocity_127', 'Duration_0.8.16',
              'Position_8', 'Program_24', 'Pitch_50', 'Velocity_127', 'Duration_0.8.16',
              'Program_24', 'Pitch_62', 'Velocity_127', 'Duration_0.8.16', 'Position_16',
              'Program_24', 'Pitch_49', 'Velocity_127', 'Duration_0.8.16', 'Program_24',
              'Pitch_61', 'Velocity_127', 'Duration_0.8.16', 'Position_24', 'Program_24',
              'Pitch_50', 'Velocity_127', 'Duration_0.8.16', 'Program_24', 'Pitch_62',
              'Velocity_127', 'Duration_0.8.16', 'Position_32', 'Program_24',
              'Pitch_40', 'Velocity_127', 'Duration_1.0.16', 'Program_24', 'Pitch_52',
              'Velocity_127', 'Duration_1.0.16', 'Program_24', 'Pitch_64', 'Velocity_127', 'Duration_1.0.16',

              'Bar_None', 'TimeSig_6/8', 'Position_0', 'Program_24', 'Pitch_52', 'Velocity_127',
              'Duration_0.8.16', 'Program_24', 'Pitch_64', 'Velocity_127', 'Duration_0.8.16',
              'Position_8', 'Program_24', 'Pitch_50', 'Velocity_127', 'Duration_0.8.16',
              'Program_24', 'Pitch_62', 'Velocity_127', 'Duration_0.8.16', 'Position_16',
              'Program_24', 'Pitch_49', 'Velocity_127', 'Duration_0.8.16', 'Program_24',
              'Pitch_61', 'Velocity_127', 'Duration_0.8.16', 'Position_24', 'Program_24',
              'Pitch_50', 'Velocity_127', 'Duration_0.8.16', 'Program_24', 'Pitch_62',
              'Velocity_127', 'Duration_0.8.16', 'Position_32', 'Program_24',
              'Pitch_40', 'Velocity_127', 'Duration_1.0.16', 'Program_24', 'Pitch_52',
              'Velocity_127', 'Duration_1.0.16', 'Program_24', 'Pitch_64', 'Velocity_127', 'Duration_1.0.16',

              'Bar_None', 'TimeSig_6/8', 'Position_0', 'Program_24', 'Pitch_52', 'Velocity_127',
              'Duration_0.8.16', 'Program_24', 'Pitch_64', 'Velocity_127', 'Duration_0.8.16',
              'Position_8', 'Program_24', 'Pitch_50', 'Velocity_127', 'Duration_0.8.16',
              'Program_24', 'Pitch_62', 'Velocity_127', 'Duration_0.8.16', 'Position_16',
              'Program_24', 'Pitch_49', 'Velocity_127', 'Duration_0.8.16', 'Program_24',
              'Pitch_61', 'Velocity_127', 'Duration_0.8.16', 'Position_24', 'Program_24',
              'Pitch_50', 'Velocity_127', 'Duration_0.8.16', 'Program_24', 'Pitch_62',
              'Velocity_127', 'Duration_0.8.16', 'Position_32', 'Program_24',
              'Pitch_40', 'Velocity_127', 'Duration_1.0.16', 'Program_24', 'Pitch_52',
              'Velocity_127', 'Duration_1.0.16', 'Program_24', 'Pitch_64', 'Velocity_127', 'Duration_1.0.16',

              'Bar_None', 'TimeSig_4/4', 'Position_0', 'Program_24', 'Pitch_52', 'Velocity_127',
              'Duration_0.8.16', 'Program_24', 'Pitch_64', 'Velocity_127', 'Duration_0.8.16',
              'Position_8', 'Program_24', 'Pitch_50', 'Velocity_127', 'Duration_0.8.16',
              'Program_24', 'Pitch_62', 'Velocity_127', 'Duration_0.8.16', 'Position_16',
              'Program_24', 'Pitch_49', 'Velocity_127', 'Duration_0.8.16', 'Program_24',
              'Pitch_61', 'Velocity_127', 'Duration_0.8.16', 'Position_24', 'Program_24',
              'Pitch_50', 'Velocity_127', 'Duration_0.8.16', 'Program_24', 'Pitch_62',
              'Velocity_127', 'Duration_0.8.16', 'Position_32', 'Program_24',
              'Pitch_40', 'Velocity_127', 'Duration_1.0.16', 'Program_24', 'Pitch_52',
              'Velocity_127', 'Duration_1.0.16', 'Program_24', 'Pitch_64', 'Velocity_127', 'Duration_1.0.16',

              'Bar_None', 'TimeSig_4/4', 'Position_0', 'Program_24', 'Pitch_52', 'Velocity_127',
              'Duration_0.8.16', 'Program_24', 'Pitch_64', 'Velocity_127', 'Duration_0.8.16',
              'Position_8', 'Program_24', 'Pitch_50', 'Velocity_127', 'Duration_0.8.16',
              'Program_24', 'Pitch_62', 'Velocity_127', 'Duration_0.8.16', 'Position_16',
              'Program_24', 'Pitch_49', 'Velocity_127', 'Duration_0.8.16', 'Program_24',
              'Pitch_61', 'Velocity_127', 'Duration_0.8.16', 'Position_24', 'Program_24',
              'Pitch_50', 'Velocity_127', 'Duration_0.8.16', 'Program_24', 'Pitch_62',
              'Velocity_127', 'Duration_0.8.16', 'Position_32', 'Program_24',
              'Pitch_40', 'Velocity_127', 'Duration_1.0.16', 'Program_24', 'Pitch_52',
              'Velocity_127', 'Duration_1.0.16', 'Program_24', 'Pitch_64', 'Velocity_127', 'Duration_1.0.16',

              ]


    idxs = [tokenizer[t] for t in tokens]
    midi = tokenizer.tokens_to_midi(idxs)
    import tempfile

    from fractions import Fraction
    # Save midi to a tempfile
    with tempfile.NamedTemporaryFile(suffix='.mid') as f:
        output_path = f.name
        midi.dump(output_path)
        # Reload with musiclang
        from musiclang import Score
        score = Score.from_midi(output_path, tokenize_before=True)
        assert [c.duration for c in score] == [Fraction(4, 1),
             Fraction(3, 1),
             Fraction(4, 1),
             Fraction(3, 1),
             Fraction(3, 1),
             Fraction(4, 1),
             Fraction(4, 1)]




"""
Copyright (c) 2023, Florian GARDIN
All rights reserved.

This source code is licensed under the BSD-style license found in the
LICENSE file in the root directory of this source tree.
"""
from musiclang import Melody


def put_on_same_chord(score):
    """Take the first chord as reference,
    Put everything into the first chord preserving the note value (It will change the piece to a static harmony)

    Parameters
    ----------
    score : Score
            Input score
        

    Returns
    -------


    """
    from musiclang import Silence
    first_chord = score[0]
    instruments = score.instruments
    parts = {ins: None for ins in instruments}

    for chord in score.chords:
        duration = chord.duration
        for instrument in instruments:
            parts[instrument] += chord.score.get(instrument, Silence(duration))

    return first_chord(**parts)


def repeat_until_duration(score, duration):
    """
    Repeat the score until score last at least duration
    returns a score exactly of length duration
    Parameters
    ----------
    score
    duration

    Returns
    -------
    """
    if score.duration < duration:
        # How much time we need to repeat at least
        nb_times = int(duration / score.duration) + 1
        score = score * nb_times
    score = score.get_score_between(0, duration)
    return score

def project_on_score(score, score2, keep_score=False):
    """Project harmonically the score onto the score2
    Algorithm : For each chord of score2 : get chords that belongs to score1 and reproject on chord of score2

    Parameters
    ----------
    score : Score
            Score to project on the chords of score2
    score2 : Score
        Score that contains the harmony
    keep_score : Score (Default value = False)
        Keep the voices of score2 ? (Default value = False)

    Returns
    -------
    new_score: Score
              The score projected on score2 chord progression

    """

    start_time = 0
    new_score = None
    #
    instruments1 = [(ins.split('__')[0], int(ins.split('__')[1])) for ins in score.instruments]

    for idx, chord2 in enumerate(score2):
        # Get all segments of chord
        end_time = start_time + chord2.duration

        subscore1 = score.get_score_between(start_time, end_time)
        if subscore1 is None:
            break

        subscore = subscore1.put_on_same_chord()
        #assert subscore.duration == chord2.duration, "Wrong duration when projecting"


        if keep_score:
            chord_score = chord2.score
            chord_score.update(subscore.score)
        else:
            chord_score = subscore.score
        new_score += chord2(**chord_score)

        start_time = end_time

    return new_score



def get_melody_between(voice, start, end, modulo=False):
    """
    Get the melody between start and end
    Parameters
    ----------
    voice
    start
    end

    Returns
    -------

    """
    time = 0
    new_voice = []
    to_break = False
    if end > voice.duration and modulo:
        # Repeat the voice until end <= voice.duration
        nb_times = int(end / voice.duration) + 1
        voice = voice * nb_times

    for note in voice:
        add_continuation = False
        note_duration = note.duration
        new_note = note.copy()
        if time >= end:
            break
        if (time < start) and (time + note_duration <= start):
            time += new_note.duration
            continue
        if time + note_duration >= end:
            new_note.duration = end - time
            to_break = True
        if time < start:
            new_note.duration = new_note.duration - (start - time)
            time += start - time
            add_continuation = True

        if add_continuation:
            from ..note import Continuation
            new_voice.append(Continuation(new_note.duration))
        else:
            new_voice.append(new_note)

        if new_note.duration < 0:
            raise Exception('Get a negative duration in get_chord_between')

        time += new_note.duration

        if to_break:
            break
    new_voice = Melody(new_voice)
    return new_voice

def get_chord_between(chord, start, end, complete_if_missing=False):
    """Get the chord with melodies that are between start and end

    Parameters
    ----------
    chord :
        chord on which to extract subchord melodies
    start :
        Start time (relative to chord)
    end :
        End time (relative to chord)
    complete_if_missing :
        If True, complete the chord with silence if the chord is not complete (Default value = False)
    Returns
    -------
    type
        chord, Chord, modified chord

    """
    new_parts = {ins: None for ins in chord.instruments}
    total_duration = end - start
    for part in chord.score.keys():
        new_voice = get_melody_between(chord.score[part], start, end)
        if complete_if_missing and new_voice.duration < total_duration:
            from ..note import Silence
            new_voice += Silence(total_duration - new_voice.duration)
        new_parts[part] = new_voice
        if complete_if_missing:
            assert new_voice.duration == total_duration, f"Wrong duration {new_voice.duration} for {part}"

    if len(new_parts.keys()) == 0:
        from ..note import Silence
        return chord(**{'piano__0': Silence(end - start)})
        # if new_voice.duration != chord.duration:
        #    from pdb import set_trace; set_trace()
    return chord(**new_parts)


def get_score_between(score, start=None, end=None):
    """Get the score with melodies that are between start and end

    Parameters
    ----------
    score :
        initial score
    start :
        Start time (relative to score) (Default value = None)
    end :
        End time (relative to score) (Default value = None)

    Returns
    -------
    type
        score, Score, modified score

    """
    start = start if start is not None else 0
    end = end if end is not None else score.duration
    new_score = None
    time = 0
    for chord in score.chords:
        chord_start = time
        chord_end = time + chord.duration
        if chord_end <= start:
            time += chord.duration
            continue
        elif chord_start >= end:
            # We are already arrived at the end of time
            break
        elif chord_end < end and chord_start >= start:
            # Perfect case we just copy the full chord
            new_score += chord.copy()
        else:
            # In all other cases we have to cut the chord
            new_start = start - time
            new_end = end - time
            new_chord = score.get_chord_between(chord, new_start, new_end)
            new_score += new_chord

        time += chord.duration

    return new_score



from musiclang.library import *
from musiclang import Score
import numpy as np

def test_parsimonious_voice_leading_up():
    chord_progression = [
        (I % II.M).h,
        (V % II.M).h['6'].o(-1),
        (I % II.M).w,
        (I % II.M).w['6'],
        (I % II.M).h['64'],
        (I % II.M).h.o(1),
        (V % II.M).set_duration(8),
        (I % II.M).h.o(1),
        (I % II.M).q['64'],
        (I % II.M).q['6'],
        (I % II.m).h,
        (V % II.mm).h['6'].o(-1),
        (I % II.m).w,
        (I % II.m).w['6'],
        (I % II.m).h['64'],
        (I % II.m).h.o(1),
        (V % II.mm)['7(sus4)'].set_duration(4),
        (V % II.mm)['7'].set_duration(4),
        (I % II.m).w.o(1),
        (V % IV.M).w,
        (IV % IV.M).w,
        (V % II.m)['7'].w
    ]
    score = Score(chord_progression).get_parsimonious_voice_leading(directions='up')
    assert np.diff([c.bass_pitch for c in score.chords]).min() >= 0


def test_parsimonious_voice_leading_down():
    chord_progression = [
        (I % II.M).h,
        (V % II.M).h['6'].o(-1),
        (I % II.M).w,
        (I % II.M).w['6'],
        (I % II.M).h['64'],
        (I % II.M).h.o(1),
        (V % II.M).set_duration(8),
        (I % II.M).h.o(1),
        (I % II.M).q['64'],
        (I % II.M).q['6'],
        (I % II.m).h,
        (V % II.mm).h['6'].o(-1),
        (I % II.m).w,
        (I % II.m).w['6'],
        (I % II.m).h['64'],
        (I % II.m).h.o(1),
        (V % II.mm)['7(sus4)'].set_duration(4),
        (V % II.mm)['7'].set_duration(4),
        (I % II.m).w.o(1),
        (V % IV.M).w,
        (IV % IV.M).w,
        (V % II.m)['7'].w
    ]
    score = Score(chord_progression).get_parsimonious_voice_leading(directions='down')
    assert np.diff([c.bass_pitch for c in score.chords]).max() <= 0

def test_parsimonious_voice_leading():
    chord_progression = [
        (I % V.M).h,
        (V % V.M),
        ]
    score = Score(chord_progression).get_parsimonious_voice_leading()
    assert score == Score([(I % V.M).h, (V % V.M)['6'].o(-1)])



from musiclang.library import *
from musiclang import Score

def test_get_drums():
    drums = (I % I.M)(s0, s1, drums_0=bd) + (I % I.M)(s0, s1) + (I % I.M)(s0, s1, drums_0=bd)
    only_drums = drums.get_instrument_names(['drums_0'])
    assert only_drums == (I % I.M)(drums_0=bd) + (I % I.M)(drums_0=r) + (I % I.M)(drums_0=bd)


def test_remove_drums():
    all_drums = (I % I.M)(s0, s1, drums_0=bd) + (I % I.M)(drums_0=bd) + (I % I.M)(s0, s1, drums_0=bd)
    drums_removed = all_drums.remove_drums()
    assert drums_removed == (I % I.M)(s0, s1) + (I % I.M)(piano__0=r, piano__1=r) + (I % I.M)(s0, s1)




from musiclang import Score
from musiclang.library import *
import tempfile
def test_start_with_silence():

    # Test with a midi file that starts with silence (anachrusis)
    score = (I% I.M)(r + s0 + s2 + s4)  + (I % I.M)(s0 + s0 + s2 + s4)
    with tempfile.NamedTemporaryFile(suffix='.mid') as f:
        output_path = f.name
        score.to_midi(output_path, time_signature=(4, 4))

        # Reload with musiclang
        score = Score.from_midi(output_path, tokenize_before=True)
        assert score[0].score['piano__0'] == r + s0 + s2 + s4
        assert score[1].score['piano__0'] == s0 + s0 + s2 + s4


import pandas as pd
import numpy as np
from miditoolkit import MidiFile
import partitura.musicanalysis as analysis
from musiclang.write.out.constants import REVERSE_INSTRUMENT_DICT



def load_score(filename, merge_tracks=True):
    """
    Load a score from a midi file and assign voices to each note

    Parameters
    ----------
    filename
    merge_tracks

    Returns
    -------
    notes: list
        List of notes with onset_beat, offset_beat, velocity, pitch, track, channel, voice
    bars: list
        List of bars with onset_beat, offset_beat
    instruments: dict
        Dictionary of instruments with channel as key and instrument name as value
    config: dict
        Dictionary of config with ticks_per_beats, bar_durations, tempo, tempos, time_signatures

    """

    import time as time_measure
    def get_end_of_bar(bar_start, time_signatures, ticks_per_beat):
        # Find time signature
        _, num, den, duration = [ts for ts in time_signatures if bar_start >= ts[0]][-1]
        return int(bar_start + duration * ticks_per_beat)

    def get_channel(idx, is_drum):
        if is_drum:
            return 9
        elif idx < 9:
            return idx
        else:
            return idx + 1

    def estimate_voices_with_index(x):
        y = x.copy()[['onset_quarter', 'duration_quarter', 'pitch']]
        y['duration_quarter'] -= 1e-1
        result = pd.Series(analysis.estimate_voices(y.to_records(), monophonic_voices=True))
        result.index = x.index  # Set the index to match the input
        return result

    m = MidiFile(filename)
    ticks_per_beat = m.ticks_per_beat
    score = []

    if len(m.time_signature_changes) > 0:
        time_signatures = [(ts.time, ts.numerator, ts.denominator, 4 * ts.numerator / ts.denominator) for ts in
                           m.time_signature_changes]
        time_signatures = list(sorted(time_signatures, key=lambda x: x[0]))
    else:
        time_signatures = [(0, 4, 4, 4)]

    # Calculate bars
    time = 0
    bars = []
    while time < m.max_tick:
        start_time = time
        end_time = get_end_of_bar(time, time_signatures, ticks_per_beat)
        bars.append((start_time / ticks_per_beat, end_time / ticks_per_beat))
        time = end_time

    programs = [(inst.program, inst.is_drum, get_channel(idx, inst.is_drum)) for idx, inst in enumerate(m.instruments)]
    instruments = {channel: REVERSE_INSTRUMENT_DICT[program] if not is_drum
    else f"drums_{program}" for program, is_drum, channel in programs}

    ## Step 1 : Get notes
    for idx, instrument in enumerate(m.instruments):
        id_instrument = f'0{idx}'[-2:]
        for idx_note, note in enumerate(instrument.notes):
            score.append({
                'onset_beat': note.start / ticks_per_beat,
                'duration_beat': (note.end - note.start) / ticks_per_beat,
                'onset_quarter': note.start / ticks_per_beat,
                'duration_quarter': (note.end - note.start) / ticks_per_beat,
                'onset_div': note.start,
                'duration_div': (note.end - note.start),
                'pitch': note.pitch,
                'voice': 0,
                'id': f"P{id_instrument}_n{idx_note}",
                'divs_pq': ticks_per_beat,
                'velocity': note.velocity,
                'track': idx,
                'channel': get_channel(idx, instrument.is_drum)
            })


    df = pd.DataFrame(score)

    def quantize_channel(df):
        ## Step 2 : Quantize
        from music21 import stream, note
        s = stream.Stream()

        for idx, row in df.iterrows():
            n = note.Note()
            n.quarterLength = row['duration_quarter']
            s.repeatInsert(n, [row['onset_quarter']])

        s.quantize((4,3), processOffsets=True, processDurations=True, inPlace=True)
        onsets = [e.offset for e in s]
        durations = [e.quarterLength for e in s]

        df['onset_quarter'] = onsets
        df['duration_quarter'] = durations
        df['onset_beat'] = df['onset_quarter']
        df['duration_beat'] = df['duration_quarter']
        df['onset_div'] = df['onset_quarter'] * ticks_per_beat
        df['duration_div'] = df['duration_quarter'] * ticks_per_beat
        return df

    df = df.sort_values(by=['onset_beat', 'pitch', 'duration_beat'])
    df = df.groupby('channel').apply(quantize_channel).reset_index(level=0, drop=True)
    df['offset_beat'] = df['onset_beat'] + df['duration_beat']
    df = df.sort_values(by=['onset_beat', 'pitch', 'duration_beat'])

    if merge_tracks:
        df['old_track'] = df['track']
        df['track'] = 0
        gpb = df.groupby('channel').apply(estimate_voices_with_index).reset_index(level=0, drop=True)
        try:
            df['voice'] = gpb
        except:
            df['voice'] = np.squeeze(gpb.values.T)
        df['track'] = df['old_track']
    else:
        df['voice'] = estimate_voices_with_index(df)

    df['track'] = df['track'] + 1
    df['voice'] = df['voice'] - 1

    tempos = [temp.tempo for temp in m.tempo_changes]

    config = {'ticks_per_beats': ticks_per_beat,
              'bar_durations': [bar[1] - bar[0] for bar in bars],
              'instruments': instruments,
              'tempo': tempos[0], 'tempos': tempos, 'time_signatures': time_signatures}

    return df, bars, instruments, config


"""
Copyright (c) 2023, Florian GARDIN
All rights reserved.

This source code is licensed under the BSD-style license found in the
LICENSE file in the root directory of this source tree.
"""
from functools import cached_property

from .constants import *
import re

class Chord:
    """
    Represents a chord in MusicLang.

    The chord is composed of :

        - One degree in ``I, II, III, IV, V, VI, VII``
        - One extension (optional) that represents figured bass in ``5,6,64,7,65,43,2,9,11,13``
        - One tonality itself composed of :

            - A degree in ``I, II, III, IV, V, VI, VII``
            - A mode in ``'M', 'm', 'mm', 'dorian, 'phrygian', 'lydian', 'mixolydian', 'aeolian', 'locrian'``
        - One score : A Dictionary where keys are part names, values are :func:`~musiclang.Melody`
        - One octave : An integer that specifies on which octave transpose the chord

    .. note:: In reality a chord in MusicLang is a "chord scale" because it always implies a tonality.

    Examples
    --------

    >>> from musiclang.library import *
    >>> score =  (I % I.M)(s0, s2)
    >>> score
    (I % I.M)(
        piano_0=s0,
        piano_1=s2)

    You can also pass a list of melodies to an instrument :

    >>> score =  (I % I.M)(violin=[s0, s2, s4.e + s2.e])
    >>> score
    (I % I.M)(
        violin__0=s0,
        violin__1=s2,
        violin__2=s4.e + s2.e)

    Finally can also pass a list of voices to an instrument :

    >>> score =  (I % I.M)(piano__0=s0 + s2, piano__2= s2 + s4)
    >>> score
     (I % I.M)(
         piano__0=s0 + s2,
         piano__2= s2 + s4)

    """

    EXCLUDED_ITEMS = ['__array_struct__']

    def __init__(self, element, extension='', tonality=None, score=None, octave=0, tags=None):
        """

        Parameters
        ----------
        element: int
        extension: int or str , optional
                 Represents the figured bass. Must be in ['','5','6','64','7','65','43','2','9','11','13']
        tonality: musiclang.Tonality, optional
                Represents the tonality of the chord
        score: dict, optional
               Represents the melodies and instruments played by this chord
        octave: int, optional
                Represents the base octave of the chord
        """
        self.element = element
        self.extension = str(extension)
        self.extension = self.normalize_extension()
        self.tonality = tonality
        self.octave = octave
        self.score = {} if score is None else score
        self.tags = set(tags) if tags is not None else set()

    def has_tag(self, tag):
        """
        Check if the tag exists for this object
        Returns a copy of the object
        Parameters
        ----------
        tag: str

        Returns
        -------
        chord: Chord
        """
        return tag in self.tags

    def add_tag(self, tag):
        """
        Add a tag to this object
        Returns a copy of the object
        Parameters
        ----------
        tag: str

        Returns
        -------
        chord: Chord
        """
        cp = self.copy()
        cp.tags.add(tag)
        return cp


    def transpose(self, interval):
        cp = self.copy()
        cp.tonality.degree += interval
        return cp

    def add_tags(self, tags):
        """
        Add several tags to the object.
        Returns a copy of the object

        Parameters
        ----------
        tags: List[str]
        tags to add

        Returns
        -------
        chord: Chord

        """
        cp = self.copy()
        cp.tags = cp.tags.union(set(tags))
        return cp

    def clear_note_tags(self):
        """
        Clear all tags from this object
        Returns a copy of the object
        Parameters
        ----------
        tag: str

        Returns
        -------
        chord: Chord
        """
        cp = self.copy()
        for k, v in cp.score.items():
            cp.score[k] = v.clear_note_tags()
        return cp

    def remove_tags(self, tags):
        """
        Remove several tags from the object.
        Returns a copy of the object

        Parameters
        ----------
        tags: List[str]

        Returns
        -------
        chord: Chord


        """
        cp = self.copy()
        cp.tags = cp.tags - set(tags)
        return cp

    def remove_tag(self, tag):
        """
        Remove a tag from this object
        Returns a copy of the object
        Parameters
        ----------
        tag: str

        Returns
        -------
        chord: Chord
        """
        cp = self.copy()
        cp.tags.remove(tag)
        return cp

    def clear_tags(self):
        """
        Clear all tags from this object
        Returns a copy of the object
        Parameters
        ----------
        tag: str

        Returns
        -------
        chord: Chord
        """
        cp = self.copy()
        cp.tags = set()
        return cp

    def has_pitch(self, pitch):
        """
        Check whether the pitch belong to the chord or not

        Parameters
        ----------
        pitch: int
        Pitch to check

        Returns
        -------
        result: boolean
            True if the chord has this pitch

        """
        return (pitch % 12) in [p % 12 for p in self.chord_pitches]

    def parse(self, pitch):
        """
        Parse an integer pitch (0=C5)
        Parameters
        ----------
        pitch :
            

        Returns
        -------

        """
        from .note import Note
        if pitch % 12 in self.scale_set:
            scale = self.scale_pitches
            type = 's'
        else:
            scale = self.chromatic_scale_pitches
            type = 'h'

        scale_mod = [s % 12 for s in scale]
        idx = scale_mod.index(pitch % 12)
        oct = (pitch - scale[0]) // 12
        return Note(type, idx, oct, 1)

    def project_on_score(self, score2, **kwargs):
        """
        Transform chord into a score and apply Score.project_on_score
        Parameters
        ----------
        score2: other score on which to project
        voice_leading:
        kwargs
        Returns
        -------
        """
        return self.to_score().project_on_score(score2, **kwargs)

    def decompose_duration(self):
        """ """
        return self(**{key: melody.decompose_duration() for key, melody in self.score.items()}, tags=self.tags)

    @property
    def degree(self):
        return self.element

    def get_note_degree_accident(self, note):
        from .constants import DEGREE_TO_SCALE_DEGREE
        if note.type == 's':
            return note.val
        elif note.type in ['l', 'r']:
            return (None, None)
        elif note.type == 'h':
            pass


    def realize_tags(self, last_note=None, final_note=None):
        new_chord_dict = {}
        last_note = last_note or {}
        final_note = final_note or {}
        for key, melody in self.score.items():
            new_chord_dict[key] = melody.realize_tags(last_note.get(key, None), final_note.get(key, None))
        return self(**new_chord_dict)


    def correct_chord_octave(self):
        """
        Correct chord octaves to minimize distance with central C

        Returns
        -------
        score: Score
        """
        from musiclang.analyze import inverse_recursive_correct_octave
        return inverse_recursive_correct_octave(self)

    def split(self, max_length):
        """
        Split the chords that are too long into smaller chords
        Parameters
        ----------
        max_length: fraction
           The maximum length of a chord

        Returns
        -------
        score: Score
           The score with shorter chords

        """
        score = None
        if self.duration <= max_length:
            return self
        nb_chords = 1 + (self.duration // max_length)
        remaining_duration = self.duration % max_length
        from musiclang.library import s0
        chords = [self(s0.augment(max_length)) for i in range(nb_chords)]
        if remaining_duration > 0:
            chords[-1].score['piano__0'] = chords[-1].score['piano__0'].set_duration(remaining_duration)
        else:
            chords = chords[:-1]
        chords = sum(chords, None)
        return self.project_on_score(chords)


    @property
    def mode(self):
        return self.tonality.mode

    def change_mode(self, mode):
        """

        Parameters
        ----------
        mode :
            

        Returns
        -------

        """
        new_chord = self.copy()
        new_chord.tonality = new_chord.tonality.change_mode(mode)
        return new_chord

    @cached_property
    def top_note(self):
        return self.extension_notes[-1]

    @cached_property
    def lowest_note(self):
        return self.extension_notes[0]

    def middle_note(self, index):
        if len(self.extension_notes) == 1:
            return self.extension_notes[0]
        if len(self.extension_notes) == 2:
            return self.extension_notes[0]
        if len(self.extension_notes) == 3:
            return self.extension_notes[1]
        if len(self.extension_notes) == 4:
            return self.extension_notes[index+1]
        else:
            return self.extension_notes[index + 1]

    @cached_property
    def middle_note1(self):
        return self.middle_note(0)

    @cached_property
    def middle_note2(self):
        return self.middle_note(1)


    @property
    def pedal(self):
        """
        Apply pedal on first note and release on last
        """
        return self(**{key: item.to_melody().pedal for key, item in self.score.items()}, tags=set(self.tags))

    def to_pitch(self, note, last_pitch=None):
        """

        Parameters
        ----------
        note :
            
        last_pitch :
             (Default value = None)

        Returns
        -------

        """
        from .out.to_midi import note_to_pitch_result
        if note.is_continuation:
            return last_pitch

        if not note.is_note:
            return None
        return note_to_pitch_result(note, self, last_pitch=last_pitch)

    def to_absolute_note(self, last_pitch=None, return_last_pitch=False):
        if last_pitch is None:
            last_pitch = {}

        if return_last_pitch:
            new_chord = {}
            for part, melody in self.score.items():
                new_melody, last_pitch[part] = melody.to_absolute_note(
                    self, last_pitch=last_pitch.get(part, None), return_last_pitch=return_last_pitch)
                new_chord[part] = new_melody
            return self(**new_chord), last_pitch
        else:
            return self(**{part: melody.to_absolute_note(self, last_pitch=last_pitch.get(part, None),
                                                     return_last_pitch=False) for part, melody in self.score.items()})


    @cached_property
    def bass_pitch(self):
        """
        Get the bass pitch of the chord

        Returns
        -------
        res: int

        """
        return self.chord_extension_pitches[0]

    def to_sequence(self):
        """
        See :func:`~Score.to_sequence()
        """
        sequence_dict = []
        for inst in self.instruments:
            sequence_dict += self.score[inst].to_sequence(self, inst)

        return sequence_dict

    def to_score(self, copy=True):
        """ """
        from .score import Score
        chord = self if not copy else self.copy()
        return Score([chord])

    def show(self, *args, **kwargs):
        """
        See :func:`~Score.show()
        """
        return self.to_score().show(*args, **kwargs)


    def to_musicxml(self, *args, **kwargs):
        return self.to_score().to_musicxml(*args, **kwargs)


    def to_music21(self, *args, **kwargs):
        return self.to_score().to_music21(*args, **kwargs)


    @cached_property
    def scale_degree(self):
        """

        Returns
        -------

        """
        tonic = self.scale_pitches[0]
        return DEGREE_TO_SCALE_DEGREE[tonic % 12] + tonic // 12

    @cached_property
    def pitch_set(self):
        """
        Get a frozenset of :func:`~Chord.chord_pitches`
        """
        return frozenset({s % 12 for s in self.chord_pitches})

    @property
    def pitch_dict(self):
        chromatic_dict = {self.to_pitch(note): note for note in self.chromatic_scale_notes}
        diatonic_dict = {self.to_pitch(note): note for note in self.scale_notes}

        return {**chromatic_dict, **diatonic_dict}


    @cached_property
    def scale_set(self):
        """
        Get a frozenset of :func:`~Chord.scale_pitches`
        """
        return frozenset({s % 12 for s in self.scale_pitches})

    @cached_property
    def chord_pitches(self):
        """
        Get the chord absolute pitches (integers) starting with the chord root.

        See Also
        --------
        :func:`~Tonality.scale_pitches`
        :func:`~Tonality.chord_extension_pitches`
        :func:`~Chord.to_pitch()`

        Returns
        -------
        res: List[int]

        Examples
        --------

        >>> from musiclang.library import *
        >>> (I % I.M).chord_pitches
        [0, 4, 7]

        >>> from musiclang.library import *
        >>> (I % I.M)['6'].chord_pitches
        [0, 4, 7]
        """
        notes = self.chord_notes
        return [self.to_pitch(n) for n in notes]

    def remove_accidents(self):
        return Chord(**{key: val.remove_accidents() for key, val in self.score.items()}, tags=set(self.tags))

    def predict_score(self, **kwargs):
        """
        Predict the continuation of the chord/score with a LLM model.
        See :func:`~Score.predict_score()`
        """
        return self.to_score().predict_score(**kwargs)

    def apply_pattern(self, chords, instruments=None, voicing=None, restart_each_chord=False,
                      fixed_bass=True, voice_leading=True, amp='mf'):
        """
        Apply a chord progression to a pattern
        Parameters
        ----------
        pattern: Score or Chord
            Pattern to use
        chords: Score or list or Chord
            Chord progression to use
        instruments: list[str] or None
            Instruments to use, by default use only piano
        voicing: list[Note] or None
            Voicing to use in the pattern. By default, spread the voicing between root -2 octave and fifth +1 octave
        restart_each_chord: bool
            If True, the pattern will be restarted at each chord
        fixed_bass: bool
            If True, the bass will be fixed in the voice leading
        voice_leading: bool
            If True, the voice leading optimizer will be applied
        amp: str
            Amplitude of the pattern (in ppp, pp, p, mp, mf, f, ff, fff)

        Returns
        -------
        Score

        """
        from musiclang.predict.composer import apply_pattern
        return apply_pattern(self, chords, instruments=instruments, voicing=voicing, restart_each_chord=restart_each_chord,
                      fixed_bass=fixed_bass, voice_leading=voice_leading, amp=amp)

    def normalize_chord_duration(self):
        """
        Normalize the duration of the chord (project on the melody with the shortest duration)
        See :func:`~Score.normalize_chord_duration()`
        """
        min_duration = min([melody.duration for melody in self.score.values()])
        base_chord = self.to_chord().set_duration(min_duration)
        return self.project_on_score(base_chord).chords[0]

    def get_scale_from_type(self, type):
        if type == "h":
            return self.chromatic_pitches
        elif type == "s":
            return self.scale_pitches
        elif type == "b":
            return self.chord_extension_pitches
        elif type == "c":
            return self.chord_pitches
        else:
            raise ValueError("This type is not associated to a scale")
    @cached_property
    def chromatic_pitches(self):
        return [self.scale_pitches[0] + i for i in range(12)]
    @cached_property
    def scale_pitches(self):
        """
        Get the scale absolute pitches (integers) starting with the chord root.

        See Also
        --------
        :func:`~Tonality.scale_pitches`
        :func:`~Chord.to_pitch()`

        Returns
        -------
        res: List[int]

        Examples
        --------

        >>> from musiclang.library import *
        >>> (I % I.M).scale_pitches
        [0, 2, 4, 5, 7, 9, 11]

        """
        from .tonality import Tonality
        if self.tonality is None:
            return (self % Tonality(0)).scale_pitches
        tonality_scale_pitches = self.tonality.scale_pitches
        start_idx = self.element
        scale_pitches = tonality_scale_pitches[start_idx:] + [t + 12 for t in tonality_scale_pitches[:start_idx]]
        scale_pitches = [n + 12 * self.octave for n in scale_pitches]
        return scale_pitches


    @cached_property
    def chromatic_scale_pitches(self):
        """ """
        return [self.scale_pitches[0] + i for i in range(12)]

    @cached_property
    def chord_extension_pitches(self):
        """
        Get the chord absolute pitches (integers) starting with the chord bass as indicated by the extension.

        See Also
        --------
        :func:`~Tonality.chord_pitches`
        :func:`~Chord.to_pitch()`

        Returns
        -------
        res: List[int]

        Examples
        --------

        >>> from musiclang.library import *
        >>> (I % I.M)['6'].chord_extension_pitches
        [4, 7, 12]

        >>> from musiclang.library import *
        >>> (I % I.M)['2[add6]'].chord_extension_pitches
        [-1, 0, 2, 4, 9]
        """

        notes = self.extension_notes
        return [self.to_pitch(n) for n in notes]

    @property
    def chord_extension_pitch_classes(self):
        return [i % 12 for i in self.chord_extension_pitches]
    @property
    def parts(self):
        """
        Get the list of all parts names
        Same as :func:`~Chord.instruments`
        """
        return list(self.score.keys())

    def to_drum(self):
        return self(**{key: val.to_drum() if key.startswith('drum') else val for key, val in self.items()})

    @property
    def instruments(self):
        """
        Get the list of all part names
        """
        return list(self.score.keys())

    def get_part(self, item):
        """
        Get the corresponding part of the chord score. Can be an integer (part index) or a string (part name)

        Parameters
        ----------
        item : str or int
               Part name or part index

        Returns
        -------
        melody: Melody
                The melody corresponding to the part

        """
        if isinstance(item, int):
            return self.score[self.parts[item]]
        else:
            return self.score[item]

    def score_equals(self, other):
        """
        Check if the parts of both chords are equal

        Parameters
        ----------
        other : Chord

        Returns
        -------
        res : bool

        """
        if not isinstance(other, Chord):
            return False
        return self.score == other.score

    def chord_equals(self, other):
        """
        Check if two chords are equal

        Parameters
        ----------
        other :
            

        Returns
        -------
        res: bool

        """
        if not isinstance(other, Chord):
            return False
        return self.element == other.element and self.extension == other.extension and self.tonality == other.tonality and self.octave == other.octave

    def __eq__(self, other):
        from .score import Score
        if isinstance(other, Score):
            return Score([self]) == other
        if not isinstance(other, Chord):
            return False
        return self.chord_equals(other) and self.score_equals(other)

    def __and__(self, other):
        if isinstance(other, int):
            return self(**{part: melody & other for part, melody in self.score.items()})
        else:
            raise Exception(f'Not compatible type with & {other.__class__}')

    def __hash__(self):
        return hash(self.__repr__())

    @property
    def possible_notes(self):
        """
        List the notes that belong to the chord (only the chord, not the scale)

        Returns
        -------
        notes: List[Note]
               The list of possible notes for a chord

        """
        from .note import Note
        if self.extension in ['5', '', '6', '64']:
            return [Note("s", 0, 0, 1), Note("s", 2, 0, 1), Note("s", 4, 0, 1)]
        elif self.extension in ['7', '65', '43', '2']:
            return [Note("s", 0, 0, 1), Note("s", 2, 0, 1), Note("s", 4, 0, 1), Note("s", 6, 0, 1)]
        elif self.extension == '9':
            return [Note("s", 0, 0, 1), Note("s", 2, 0, 1), Note("s", 4, 0, 1), Note("s", 6, 0, 1), Note("s", 1, 1, 1)]
        elif self.extension == '11':
            return [Note("s", 0, 0, 1), Note("s", 2, 0, 1), Note("s", 4, 0, 1), Note("s", 6, 0, 1), Note("s", 1, 1, 1),
                    Note("s", 3, 1, 1)]
        elif self.extension == '13':
            return [Note("s", 0, 0, 1), Note("s", 2, 0, 1), Note("s", 4, 0, 1), Note("s", 6, 0, 1),
                    Note("s", 1, 1, 1), Note("s", 3, 1, 1), Note("s", 5, 1, 1)]
        else:
            raise Exception('Unknown extension : ' + self.extension)

    @property
    def scale_notes(self):
        """
        List the notes that belong to the chord scale

        Returns
        -------
        notes: List[Note]
               The list of possible notes for the chord scale
        """
        from .note import Note
        return [Note("s", i, 0, 1) for i in range(7)]

    @property
    def chromatic_scale_notes(self):
        """
        List the chromatic notes that belong to the chord scale

        Returns
        -------
        notes: List[Note]
               The list of possible chromatic notes for the chord scale
        """
        from .note import Note
        return [Note("h", i, 0, 1) for i in range(12)]

    @property
    def scale_dissonances(self):
        """
        Returns the notes that don't belong to the chord but belong to the chord scale

        Returns
        -------
        notes: List[Note]
               The list of possible notes for the chord scale
        """
        consonances = {(note.val % 7) for note in self.possible_notes}
        return [note for note in self.scale_notes if note.val not in consonances]

    def __getitem__(self, item):
        """
        Attach a figured bass to the chord (eg : 64)

        Parameters
        ----------

        item: int or str in ['5', '7', '9', '11', '13', '65', '43', '2', '6', '64', '']
              Extension of the chord

        Returns
        -------

        chord: Chord
               A new chord with the figured bass extension

        """
        item = str(item)
        res = self.copy()
        res.extension = str(item)
        try:
            n = res.extension_notes
        except Exception as e:
            raise e
            #raise ValueError(f'Could not parse extension : {item} {e}')
        res.extension = res.normalize_extension()

        return res

    #
    # def __getattr__(self, item):
    #     chord = self.copy()
    #     chord.score = {k: getattr(s, item) for k, s in self.score.items()}
    #     return chord

    @property
    def duration(self):
        """
        Get the chord duration as the max duration of the melodies contained in the chords

        Returns
        -------
        duration: fractions.Fraction or int
                  duration of the chord
        """
        if len(self.score.keys()) == 0:
            return 0
        return max([self.score[key].duration for key in self.score.keys()])

    def set_amp(self, amp):
        return self(**{key: val.set_amp(amp) for key, val in self.items()}, tags=set(self.tags))

    def set_part(self, part, melody, inplace=False):
        """
        Set a part on the chord (inplace operation)
        Parameters
        ----------
        part : part name

        melody :  melody to apply to the part
        inplace: bool
                 If the operation is inplace

        Returns
        -------

        """

        if inplace:
            score = self
        else:
            score = self.copy()
        part_name = score.parts[part] if isinstance(part, int) else part
        score[part_name] = melody
        return score


    def has_score(self):
        """

        """
        return self.score is not None

    def to_chord(self, duration=False):
        """
        Returns a copy of the chord without the chord score

        Returns
        -------
        chord: Chord
               Chord with empty score

        """
        result = self.copy()
        result.score = {}
        if duration:
            result = result.set_duration(self.duration)
        return result

    @property
    def chords(self):
        """
        Returns a Score object with this chord

        Returns
        -------
        score: Score
               The score with this chord as the only element

        """
        from .score import Score
        return Score([self]).chords

    def __mul__(self, other):
        """
        If other is Integer, repeat the note other times
        """
        from .score import Score
        if isinstance(other, int):
            return Score([self.copy() for i in range(other)])
        else:
            raise Exception('Cannot multiply Chord and ' + str(type(other)))

    def __mod__(self, other):
        """
        Modulate to another tonality relative to the current tonality

        Parameters
        ----------
        other: Tonality
            Tonality on which to modulate this chord
        """
        from .tonality import Tonality
        if isinstance(other, Tonality):
            chord = self.copy()
            # Tonality is always referenced from base tonality which is the most outside tonality
            other = other.copy()
            other.octave += self.octave
            chord.tonality = chord.tonality + other
            chord.octave = 0
            return chord
        else:
            raise Exception('Following % should be a Tonality')

    def modulate(self, other):
        """
        Modulate to another tonality relative to the current tonality

        Parameters
        ----------
        other: Tonality
            Tonality on which to modulate this chord
            

        Returns
        -------
        chord: Chord
               Modulated chord

        """
        from .tonality import Tonality
        if isinstance(other, Tonality):
            chord = self.copy()
            # Tonality is always referenced from base tonality which is the most outside tonality
            other = other.copy()
            other.octave += self.octave
            chord.tonality = chord.tonality + other
            chord.octave = 0
            return chord
        else:
            raise Exception('Following % should be a Tonality')

    def __add__(self, other):
        """
        Add a new chord to this chord, it will returns a score

        Parameters
        ----------

        other: Chord or Score
               Create a new score concatenating this and the other object

        Examples
        ---------

        >>> from musiclang.library import *
        >>> chord1 = (I % I.M)(piano__0=s0)
        >>> chord2 = (V % I.M)(piano__0=s0)
        >>> chord + chord2
        (I % I.M)(piano__0=s0) + (V % I.M)(piano__0=s0)

        """

        from .score import Score
        if other is None:
            return self.copy()
        if isinstance(other, Chord):
            return Score([self.copy(), other.copy()])
        if isinstance(other, Score):
            return Score([self.copy()] + other.copy().chords)

    def __radd__(self, other):
        from .score import Score
        if other is None:
            return Score([self.copy()])

    def element_to_str(self):
        """
        Convert the element of the chord into a string

        Returns
        -------
        res: str

        """
        return ELEMENT_TO_STR[self.element]

    def extension_to_str(self):
        """
        Converts the extension of the chord into a string

        Returns
        -------
        res: str

        """
        if str(self.extension) == '5' or str(self.extension) == '':
            return ''
        else:
            return f"['{self.extension}']"


    def tonality_to_str(self):
        """
        Convert the tonality of the chord to a string.

        Parameters
        ----------

        Returns
        -------

        """
        if self.tonality is None:
            return ""

        return " % " + self.tonality.to_code()

    def d(self):
        """Chord up one octave (=o(-1))"""
        return self.o(-1)

    def u(self):
        """Chord down one octave (=o(-1))"""
        return self.o(1)

    def to_extension_note(self):
        return self(**{key: val.to_extension_note(self) for key, val in self.items()}, tags=set(self.tags))

    def to_chord_note(self):
        return self(**{key: val.to_chord_note(self) for key, val in self.items()}, tags=set(self.tags))

    def to_standard_note(self):
        return self(**{key: val.to_standard_note(self) for key, val in self.items()}, tags=set(self.tags))


    def o_melody(self, octave):
        """Change the octave of the melody (not the chord)

        Parameters
        ----------
        octave : int,
                Number of octaves

        Returns
        -------
        chord : Chord

        """
        new_parts = {}
        for part in self.score.keys():
            new_parts[part] = self.score[part].o(octave)

        return self(**new_parts).add_tags(self.tags)

    def patternize(self,
                 nb_excluded_instruments=0,
                 fixed_bass=True,
                 voice_leading=True,
                 melody=False,
                 instruments=None,
                 voicing=None,
                 add_metadata=False,
                   max_duration=64,
                   **kwargs):
        """
        Extract the pattern from the chord

        Parameters
        ----------
        nb_excluded_instruments: int (Default value = 0)
            Number of instruments to exclude from the pattern
        fixed_bass: bool (Default value = True)
            If True, the bass will be fixed in the pattern response
        voice_leading: bool (Default value = True)
            If True, the voice leading will be applied in the pattern response
        melody: bool (Default value = False)
            Do you want to extract a melody or an accompaniment pattern
        instruments: list[str] (Default value = None)
            List of instruments to use for the pattern
        voicing: list[Note] or None (Default value = None)
            Voicing to use for the pattern, otherwise extract the good one
        add_metadata: bool (Default value = True)
            If True, add metadata and features to the pattern
        max_duration: int (Default value = 64)
            Maximum duration of the chord to apply patternize
        Returns
        -------
        score_pattern: Chord
            Patternized chord or score
        pattern: dict
            Pattern data of the score

        """
        from musiclang.analyze.pattern_analyzer import PatternExtractor
        if self.duration > max_duration:
            raise Exception(f'Patternize only works on chords with duration <= {max_duration} increase the max_duration parameter if you want to use it on longer chords')

        dict_pattern = PatternExtractor(
                 nb_excluded_instruments=nb_excluded_instruments,
                 fixed_bass=fixed_bass,
                 voice_leading=voice_leading,
                 melody=melody,
                 instruments=instruments,
                 voicing=voicing
                 ).extract(self)

        pattern = dict_pattern['orchestra']['pattern']
        if add_metadata:
            from musiclang.analyze.pattern_analyzer import PatternFeatureExtractor
            dict_pattern = PatternFeatureExtractor().extract(dict_pattern,
                                                             self,
                                                             melody=melody,
                                                             nb_excluded_instruments=nb_excluded_instruments,
                                                             )
        dict_pattern['multi_chord'] = False
        return dict_pattern, pattern

    def project_pattern(self, score, restart_each_chord=False):
        return self.to_score().project_pattern(score, restart_each_chord=restart_each_chord)

    def remove_drums(self):
        return self(**{ins: val for ins, val in self.score.items() if not ins.startswith('drums')})


    def o(self, octave):
        """Chord up or down the amount of octave in parameter, it will change the chord octave, not the melody

        Parameters
        ----------
        octave :
            return:

        Returns
        -------

        """
        c = self.copy()
        c.octave += octave
        return c


    def delete_instruments(self, instruments):
        return self(**{key: val for key, val in self.score.items() if key not in instruments})

    def delete_instrument_names(self, instrument_names):
        return self(**{key: val for key, val in self.score.items() if key.split('__')[0] not in instrument_names})

    def get_chord_between(self, start, end, complete_if_missing=False):
        """
        Parameters
        ----------
        start :
            
        end :

        complete_if_missing:
            (Default value = False)
        Complete chord with silence if missing time in chord

        Returns
        -------

        """
        from .time_utils import get_chord_between
        return get_chord_between(self, start, end, complete_if_missing=complete_if_missing)

    def copy(self):
        """
        Copy the current chord
        """
        return Chord(element=self.element,
                     extension=self.extension,
                     tonality=self.tonality.copy() if self.tonality is not None else None,
                     score={k: s.copy() for k, s in self.score.items() if s is not None} if self.score is not None else None,
                     octave=self.octave,
                     tags=set(self.tags)
                     )




    def preparse_named_melodies(self, named_melodies):
        """
        Helper function to deal with dictionary of melodies that can contain a list of melodies for some parts.
        It will automatically convert drums melodies to drums track

        Parameters
        ----------
        named_melodies : dict
                         The dictionary of named melodies
            

        Returns
        -------
        res: dict
             The

        Examples
        --------
        >>> from musiclang.library import *
        >>> from musiclang import Chord
        >>> D = {"piano":[s0, s2, s4.e + s2.e]}
        >>> Chord.preparse_named_melodies(D)
        {
            piano__0:s0,
            piano__1:s2,
            piano__2:s4.e + s2.e
            }

        """
        from musiclang import Melody

        named_melodies_result = {}
        for key in named_melodies.keys():
            key_obj = key.split('__')
            if len(key_obj) == 1:
                key_obj = key_obj[0]
                number = 0
            else:
                key_obj, number = key_obj[0], key_obj[1]
                number = int(number)

            if (isinstance(named_melodies[key], list)):
                for idx, melody in enumerate(named_melodies[key]):
                    mel = melody.to_melody()
                    if key_obj.startswith('drums'):
                        new_mel = None
                        for n in mel.notes:
                            new_mel += n.convert_to_drum_note(self)
                        mel = new_mel
                    named_melodies_result[key_obj + '__' + str(number + idx)] = mel
            else:
                mel = named_melodies[key].to_melody()
                if key_obj.startswith('drums'):
                    new_mel = None
                    for n in mel.notes:
                        new_mel += n.convert_to_drum_note(self)
                    mel = new_mel
                named_melodies_result[key_obj + '__' + str(number)] = mel

        return named_melodies_result

    def items(self):
        return self.score.items()

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, d):
        self.__dict__ = d

    def augment(self, duration):
        from musiclang import Silence
        if self.empty_score:
            silence = Silence(duration)
            return self(silence)
        else:
            return self(**{part: melody.augment(duration) for part, melody in self.score.items()}, tags=self.tags)

    def set_duration(self, duration):
        from musiclang import Silence
        if self.empty_score:
            if isinstance(duration, float):
                duration = frac(duration).limit_denominator(8)
            silence = Silence(duration)
            return self(silence)
        else:
            return self(**{part: melody.set_duration(duration) for part, melody in self.score.items()}, tags=self.tags)

    def __getattr__(self, item):
        if item in STR_TO_DURATION.keys() and self.empty_score:
            from musiclang import Silence
            duration = STR_TO_DURATION[item]
            silence = Silence(duration)
            return self(silence)

        if len(self.parts) == 0:
            raise AttributeError()
        try:
            res = self(**{part: getattr(self.score[part], item) for part in self.parts}, tags=self.tags)
            return res
        except Exception:
            raise AttributeError(f"Not existing attribute {item}")

    def __call__(self, *melodies, tags=None, **named_melodies):
        """
        Method that allows to assign melodies and instruments to a chord

        Parameters
        ----------
        *melodies: List[Melody]
                   Melodies that will be assigned to default instrument (piano)
        **named_melodies: Dict[str, Melody]
                        Dictionary of named melodies that will be unpacked if necessary

        Returns
        -------

        Examples
        ________

        >>> from musiclang.library import *
        >>> score =  (I % I.M)(s0, s2)
        >>> score
        (I % I.M)(
            piano_0=s0,
            piano_1=s2)

        You can also pass a list of melodies to an instrument :

        >>> score =  (I % I.M)(violin=[s0, s2, s4.e + s2.e])
        >>> score
        (I % I.M)(
            violin__0=s0,
            violin__1=s2,
            violin__2=s4.e + s2.e)

        Finally can also pass a list of voices to an instrument :

        >>> score =  (I % I.M)(piano__0=s0 + s2, piano__2= s2 + s4)
        >>> score
         (I % I.M)(
             piano__0=s0 + s2,
             piano__2= s2 + s4)
        """
        chord = self.copy()
        if tags is not None:
            chord = chord.add_tags(tags)
        named_melodies_preparsed = self.preparse_named_melodies(named_melodies)
        chord.score = {**{f'piano__{i}': melody.to_melody() for i, melody in enumerate(melodies)}, **named_melodies_preparsed}
        return chord

    def melody_to_str(self):
        """
        Convert The chord score to its string representation

        Returns
        -------
        res: str

        """
        return '\n' + ', \n'.join([f"\t{k}=" + str(melody) for k, melody in self.score.items()])

    def __repr__(self):
        return f"{self.to_code()}({self.melody_to_str()})"

    @property
    def full_octave(self):
        return self.octave + self.tonality.octave

    def get_inversion_index(self):
        """
        Get positive inversion index of the chord (chordal transposition, counted positive)

        Returns
        -------
        res: int

        """
        from musiclang.library import BASE_CHORDAL_TRANSLATION_DICT
        extension, replacements, additions, removals = self.get_extension_properties()
        base_transposition = BASE_CHORDAL_TRANSLATION_DICT[extension]
        return base_transposition

    def chord_distance(self, c2):
        """
        Get the chord distance between two chords

        Parameters
        ----------
        c2

        Returns
        -------
        res: tuple[int, int]
                transposition number, chordal_inversion number
                transposition number is the number of semitones between the two chords
                chordal_inversion number is the number of chordal inversion between the two chords
        """

        bass_pitch1 = self.root_pitch
        bass_pitch2 = c2.root_pitch

        inv1 = self.get_inversion_index()
        inv2 = c2.get_inversion_index()

        transposition = (bass_pitch2 - bass_pitch1)
        chordal_inversion = (inv2 - inv1)

        return transposition, chordal_inversion

    def get_parsimonious_voice_leading(self, candidate, direction=None):
        """
        Get the parsimonious voice leading between two chords

        Parameters
        ----------
        candidate: Chord
        direction: str or None (Default value = None), in ['up', 'down', None]
            Direction of the voice leading

        Returns
        -------
        best_chordal_inversion, best_octave:
            best_chordal_inversion: int
                The best chordal inversion
            best_octave: int
                The best octave


        """
        nb_notes = len(self.chord_notes)
        nb_notes_candidate = len(candidate.chord_notes)  # FIXME : handle different number of notes
        root = self.bass_pitch

        final_chord = candidate.copy()
        final_chord.octave = 0
        final_chord.tonality.octave = 0
        final_chord = final_chord.to_root_extension()

        other_root = final_chord.bass_pitch

        transposition = other_root - root
        offset_octave = round(transposition/12)
        normalized_transposition = transposition - offset_octave * 12
        final_chord = final_chord.o(-offset_octave)

        # FIXME : nb_notes_candidate here ? Or more complex formula depending on nb_notes also ?
        theta = (nb_notes_candidate * normalized_transposition) / 12  # Angular motion (divided by 2pi)
        optimal_chordal_transposition = -round(theta)

        final_chord = final_chord.invert(optimal_chordal_transposition)
        if optimal_chordal_transposition < 0:
            final_chord = final_chord.o(-1)

        if final_chord.bass_pitch > root and direction == 'down':
            # Invert the chord
            inversion_idx = final_chord.get_inversion_index()
            final_inversion_idx = inversion_idx - 1
            final_chord = final_chord.invert(final_inversion_idx - inversion_idx)
            if final_inversion_idx < 0:
                final_chord = final_chord.o(-1)
        elif final_chord.bass_pitch < root and direction == 'up':
            # Invert the chord
            inversion_idx = final_chord.get_inversion_index()
            final_inversion_idx = inversion_idx + 1
            final_chord = final_chord.invert(final_inversion_idx - inversion_idx)
            if final_inversion_idx > nb_notes_candidate - 1:
                final_chord = final_chord.o(1)

        return final_chord


    @property
    def root_pitch(self):
        """
        Get the root pitch of the chord

        Returns
        -------
        res: int

        """
        return self.scale_pitches[0]



    def replace_instruments_names(self, **instruments_dict):
        """
        Replace any instrument with another (use part name (eg: piano__0)

        Parameters
        ----------
        instruments_dict: dict[str, str]
            Dictionary of parts name to replace

        Returns
        -------
        chord: Chord

        """
        new_chord_dict = {}
        idx_instruments = {}
        instruments = sorted(self.score.keys(), key=lambda x: (x.split('__')[0], int(x.split('__')[1])))
        for ins in instruments:
            name, idx = ins.split('__')
            idx = int(idx)
            new_name = instruments_dict.get(name, name)
            if new_name not in idx_instruments.keys():
                idx_instruments[new_name] = -1
            chosen_idx = idx_instruments[new_name] + 1 if idx <= idx_instruments[new_name] else idx
            new_chord_dict[new_name + '__' + str(chosen_idx)] = self.score[ins]
            idx_instruments[new_name] = max(idx, idx_instruments[new_name])

        return self(**new_chord_dict)

    def replace_instruments(self, **instruments_dict):
        """
        Replace any instrument with another (use full part name (eg: piano__0)

        Parameters
        ----------
        instruments_dict: dict[str, str]
            Dictionary of parts name to replace

        Returns
        -------
        chord: Chord

        """
        instruments_dict = {key: instruments_dict[key] if key in instruments_dict.keys() else key for key in self.parts}
        new_chord_dict = {}
        for ins_name, new_ins_name in instruments_dict.items():
            if ins_name in self.score.keys():
                new_chord_dict[new_ins_name] = self.score[ins_name]

        return self(**new_chord_dict)


    def normalize_extension(self):
        extension, replacements, additions, removals = self.get_extension_properties()
        return self.properties_to_extension(extension, replacements, additions, removals)

    def invert(self, inversion):
        extension, replacements, additions, removals = self.get_extension_properties()
        four_notes_extensions = ['7', '65', '43', '2']
        three_notes_extensions = ['', '6', '64']

        if extension in four_notes_extensions:
            index_extension = four_notes_extensions.index(extension)
            new_extension = four_notes_extensions[(index_extension + inversion) % len(four_notes_extensions)]
        elif extension in three_notes_extensions:
            index_extension = three_notes_extensions.index(extension)
            new_extension = three_notes_extensions[(index_extension + inversion) % len(three_notes_extensions)]
        else:
            raise Exception('Unknown extension : ' + extension, '. Only three and four notes chords are supported')

        return self[self.properties_to_extension(new_extension, replacements, additions, removals)]


    def properties_to_extension(self, extension, replacements, additions, removals):
        replacements = ''.join([f'({r})' for r in replacements])
        additions = ''.join([f'[{a}]' for a in additions])
        removals = ''.join(['{' + r + '}' for r in removals])
        return extension + replacements + additions + removals

    def get_extension_properties(self):
        extension = self.extension.split('|')[0]
        replacements = sorted(re.findall(r'\((.*?)\)', extension))
        additions = sorted(re.findall(r'\[(.*?)\]', extension))
        removals = sorted(re.findall(r'\{(.*?)\}', extension))
        ext = extension
        for r in replacements + additions + removals:
            ext = ext.replace(r, '')
        ext = ext.replace('()', '').replace('[]', '').replace('{}', '')
        extension = ext
        return extension, replacements, additions, removals


    def to_root_extension(self):
        """
        Get the root extension of the chord (No bass inversion but still all replacements, additions and removals

        Returns
        -------
        chord: Chord
            The chord in root extension

        """
        extension, replacements, additions, removals = self.get_extension_properties()
        if extension in ['', '6', '64']:
            return self[self.properties_to_extension('', replacements, additions, removals)]
        elif extension in ['7', '65', '43', '2']:
            return self[self.properties_to_extension('7', replacements, additions, removals)]
        else:  # Other are necessarily in root position
            return self[self.properties_to_extension(extension, replacements, additions, removals)]

    def set_extension_text(self, new_extension):
        """
        Set the inversion text of the chord

        Parameters
        ----------
        inversion: str
            The inversion text

        Returns
        -------
        chord: Chord
            The chord with the inversion text

        """
        extension, replacements, additions, removals = self.get_extension_properties()
        return self.properties_to_extension(new_extension, replacements, additions, removals)

    def _chord_notes_calc(self, extension, replacements, additions, removals):
        from .library import BASE_EXTENSION_DICT, \
            DICT_REPLACEMENT, DICT_ADDITION, DICT_REMOVAL
        notes = BASE_EXTENSION_DICT[extension][:]
        notes_without_octave = [n.o(-n.octave) for n in notes]

        dict_replaced = {}
        for replacement in replacements:
            note_replaced, new_note = DICT_REPLACEMENT[replacement]
            if note_replaced not in notes_without_octave:
                additions.append(replacement)
            else:
                idx = notes_without_octave.index(note_replaced)
                notes[idx] = new_note.o(notes[idx].octave)
                notes_without_octave[idx] = new_note.o(-new_note.octave)
                dict_replaced[note_replaced] = notes_without_octave[idx]

        for addition in additions:
            note_after, new_note = DICT_ADDITION[addition]
            if note_after in dict_replaced.keys():
                query_note = dict_replaced[note_after]
            else:
                query_note = note_after
            idx = notes_without_octave.index(query_note) + 1
            new_note = new_note.o(note_after.octave)
            notes.insert(idx, new_note)
            notes_without_octave.insert(idx, new_note.o(-new_note.octave))

        for removal in removals:
            note_removed = DICT_REMOVAL[removal]
            idx = len(notes) - notes_without_octave[::-1].index(note_removed) - 1
            notes.pop(idx)
            notes_without_octave.pop(idx)
        # Sort per pitch
        notes = list(sorted(notes, key=lambda x: self.to_pitch(x)))
        return notes

    @property
    def chord_notes(self):
        extension, replacements, additions, removals = self.get_extension_properties()
        new_extension = ''
        if extension in ['2', '65', '43', '7']:
            new_extension = '7'
        elif extension == '9':
            new_extension = '9'
        elif extension == '11':
            new_extension = '11'
        elif extension == '13':
            new_extension = '13'

        return self._chord_notes_calc(new_extension, replacements, additions, removals)

    @cached_property
    def extension_notes(self):
        # Try a composite with a match
        extension, replacements, additions, removals = self.get_extension_properties()
        return self._chord_notes_calc(extension, replacements, additions, removals)


    def optimal_projection(self, chord2):
        """
        Get the optimal projection of the chord on the other chord, it will minimize the movement of voices while
        preserving chord notes to chord notes

        Parameters
        ----------
        chord2: Other chord

        Returns
        -------
        chord: Chord

        """
        chord3 = self.get_parsimonious_voice_leading(chord2)
        octave, trans_chord = chord2.chord_distance(chord3)
        pattern = self.to_pattern()
        for pat in pattern:
            octave = octave // 12
            pat['voice'] = pat['voice'].add_value_chord(trans_chord, octave, len(chord2.chord_notes))
            # pat['voice'] = pat['voice'].o(octave)
        from musiclang import Score
        return Score.from_pattern(pattern, [chord2], use_voice=True).chords[0]


    def to_voicing(self, nb_voices=4, instruments=None):
        """Convert score to a four voice voicing using the extensions provided in the chord.

        It will remove the existing scores of each chord and create the associated voicings

        Parameters
        ----------
        instruments : None or List[str] (Default value = ['piano__0', 'piano__1', 'piano__2', 'piano__3'])
                      The list of instruments used to create the voicings.

        Returns
        -------

        score: Score
               The score with voicings corresponding to chords

        """

        if instruments is None:
            instruments = [f'piano__{i}' for i in range(nb_voices)]

        notes = self.extension_notes[:nb_voices]
        if not self.empty_score:
            notes = [n.set_duration(self.duration) for n in notes]
        if len(notes) < nb_voices:
            # Add to match octaves
            for i in range(len(notes), nb_voices):
                n = notes[i % len(notes)].copy()
                oct = (i // len(notes))
                n.octave += oct
                notes.append(n)
        chord = self(**{ins: notes[i].copy() for i, ins in enumerate(instruments)})

        return chord


    @property
    def empty_score(self):
        return len(self.score.keys()) == 0

    def to_code(self):
        """Export the code as python interpretable str.
        It ignores the score
        :return:

        Parameters
        ----------

        Returns
        -------

        """
        chord_str = f"({self.element_to_str()}{self.extension_to_str()}{self.tonality_to_str()})"
        if self.octave != 0:
            chord_str = f"{chord_str}.o({self.octave})"

        return chord_str

    def to_midi(self, filepath, **kwargs):
        """
        Save the chord to midi

        See Also
        --------

        :func:`~Score.to_midi()`

        Parameters
        ----------
        filepath :
            
        **kwargs :
            

        Returns
        -------

        """
        from .score import Score
        return Score([self]).to_midi(filepath, **kwargs)


    def to_scale_notes(self):
        """
        Convert the notes of a chord in scale notes

        Parameters
        ----------

        Returns
        -------

        """
        return self(**{ins: n.to_scale_notes(self) for ins, n in self.score.items()})

    def project_on_rhythm(self, rhythm, **kwargs):
        return self(**{ins: n.project_on_rhythm(rhythm, chord=self, **kwargs) for ins, n in self.score.items()})

    def set_degree(self, degree):
        res = self.copy()
        res.element = degree
        return res

    def set_octave(self, octave):
        res = self.copy()
        res.octave = octave
        return res
    def set_element(self, element):
        return self.set_degree(element)

    def to_custom_chords(self, nb_voices=4):
        return self.to_score().to_custom_chords(nb_voices=nb_voices)


    def to_pattern(self, drop_drums=False, nb=4, pattern=True):
        """
        Convert the chord to a pattern (like a musical template)

        Parameters
        ----------
        drop_drums: bool (Default value = True)
            If True, drop the drums instruments
        nb: int (Default value = 4)
            nb arpeggio notes to consider on each side, it will give the acceptable range of the instrument melody.
            You usually don't need it because it is used only for the musical grid
        pattern: bool (Default value = True)
            If True, return the pattern inside the response

        Returns
        -------
        pattern: Pattern

        """
        from musiclang.transform.composing.chord_to_orchestra import chord_to_pattern
        return chord_to_pattern(self, drop_drums=drop_drums, nb=nb, keep_pattern=pattern)

    def repeated_notes_to_legato(self):
        return self(**{ins: n.repeated_notes_to_legato() for ins, n in self.score.items()})

    def get_raw_pattern(self):
        from .melody import Melody
        def voicing_to_quality(voicing, chord):
            notes = chord.extension_notes
            length = len(notes)
            length = min(4, length)
            qualities = []

            data_dict = {
                (0, 1): 'lowest_note',
                (0, 2): 'lowest_note',
                (0, 3): 'lowest_note',
                (0, 4): 'lowest_note',

                (1, 1): 'lowest_note',
                (1, 2): 'top_note',
                (1, 3): 'middle_note1',
                (1, 4): 'middle_note1',

                (2, 1): 'lowest_note',
                (2, 2): 'lowest_note',
                (2, 3): 'top_note',
                (2, 4): 'middle_note2',

                (3, 1): 'lowest_note',
                (3, 2): 'top_note',
                (3, 3): 'lowest_note',
                (3, 4): 'top_note',

            }
            for note in voicing:
                octave = note.octave
                value = note.val
                quality = data_dict[(value, length)]
                qualities.append((quality, octave))

            return qualities

        orchestration = []
        data, pattern = self.patternize()
        instruments = data['instruments']
        instrument_idxs = data['instrument_idxs']
        voicing = data['voicing']
        chord = data['chord']
        qualities = voicing_to_quality(voicing, data['orchestra']['pattern'])


        instruments_idx = {}
        for idx, (ins, quality, voice, part) in enumerate(zip(instruments, qualities, voicing, instrument_idxs)):
            note, octave = quality
            local_pattern = pattern.score[f'v__{idx}']
            local_pattern = Melody([n if n.type != 'x' else n.set_val(0) for n in local_pattern.notes])
            rhythm = local_pattern.quantize_melody()
            instruments_idx[ins] = instruments_idx.get(ins, -1) + 1
            orchestration.append({'instrument': ins, 'octave': octave,
                                  'note': note, 'pattern': local_pattern,
                                  'rhythm': rhythm, 'voice': voice,
                                  'part': part,
                                  'chord': chord
                                  })

        return orchestration

    @classmethod
    def from_romantext(self, text, time_signature, tonality):
        from musiclang import ScoreFormatter
        full_text = f'Time Signature: {"/".join([str(t) for t in time_signature])}'
        full_text += f'\nTonality: {tonality}'
        full_text += f'\nm0 {text}'
        return ScoreFormatter(full_text).parse()[0]

    def to_romantext(self, tonality):

        def to_degree_type(degree, mode):
            L = {
                'm': ['i', 'ii%', 'III+', 'iv', 'V', 'VI', 'viio'],
                'M': ['I', 'ii', 'iii', 'IV', 'V', 'vi', 'vii%']
            }
            return L[mode][degree]

        tone = tonality.degree
        mode = tonality.mode

        extension_tonality = ''
        if mode != self.tonality.mode or tone != self.tonality.degree:
            extension_tonality = f"/{self.tonality.to_romantext(tonality)}"

        # Now find the degree
        degree = to_degree_type(self.degree, self.tonality.mode)

        # Find proper extension
        extension = self.extension

        # Regex to list everything between "{}" and everything between "[]"
        import re
        reg_omit = r"\{(.*?)\}"
        reg_add = r"\[(.*?)\]"

        matches_omit = re.findall(reg_omit, extension)
        matches_add = re.findall(reg_add, extension)

        to_remove = ["{" + f"{match}" + "}" for match in matches_omit]
        to_remove += ["[" + f"{match}" + "]" for match in matches_add]

        formatted_extension = extension
        for r in to_remove:
            formatted_extension = formatted_extension.replace(r, '')

        # Write a script that matches the text which is not between {} or []
        # For example "64{-5}[add6][add2]" should return ["64"]

        for match in matches_omit:
            formatted_extension += f"[{match.replace('-', 'no')}]"

        for match in matches_add:
            formatted_extension += f"[{match}]"

        formatted_extension = formatted_extension.replace('(+)', '+')

        return f"{degree}{formatted_extension}{extension_tonality}"

